(**********************************************************************)
(* (c) Greg Morrisett, Neal Glew, David Walker,                       *)
(*     June 1998, all rights reserved.                                *)
(**********************************************************************)

(* cogen_main.ml
 * Generate the main program for a KML source.
 *)

open Utilities;;
open Variable;;
open Tal;;

(* For now it assume one module with one export of type int->int *)

(* warning: this is a hack *)

(* Helper functions *)

let mk_string l s : data_block =
  (l,
   Some (Cprim (PCbytearray Byte1)),
   ([Dtag (String.length s);Dbytes (Raw, s)],[Toarray]))
;;

let usage_msg = label_of_string "usage_msg";;
let uncaught_exn_msg = label_of_string "uncaught_exn_msg";;

let usage_msg_db = mk_string usage_msg "usage: <num>\n";;
let uncaught_exn_msg_db =
  mk_string uncaught_exn_msg "\nuncaught exception!\n"
;;

let call_stdlib l stack args =
  let args_size = 4*(List.length args) in
  List.fold_left (fun is a -> (Push a)::is)
    [Call (Addr l,[Tapp stack]); ArithBin (Add,Reg Esp,Immed args_size)] args
;;

let call_kml l stack1 stack2 args =
  List.fold_left (fun is a -> (Push a)::is)
    [Call (Addr l,[Tapp stack2;Tapp stack1])] args
;;

let call_closure reg stack1 stack2 args =
  let a = newvar () in
  (Unpack (a,reg,(Reg reg,[])))::
  (List.fold_left (fun is a -> (Push a)::is)
     [Call (Prjr ((reg,[]),0),[Tapp stack2;Tapp stack1])]
     (args@[(Prjr ((reg,[]),4),[])]))
;;    

let l_args = label_of_string "_args";;
let l_exit = label_of_string "_exit";;
let loop = label_of_string "inf_loop";;
let l_int_of_string = label_of_string "_int_of_string";;
let l_print_int = label_of_string "_print_int";;
let l_print_newline = label_of_string "_print_newline";;
let l_print_string = label_of_string "_print_string";;
let l_tal_main = label_of_string "_tal_main";;
let uncaught_exn = label_of_string "uncaught_exn";;
let usage = label_of_string "usage";;

let usage_cb : code_block =
  let rho = newvar () in
  let rs =
    Dict.singleton compare Esp (Capp (Cprim PCstackptr,Cvar rho)) in
  let c =
    Cforall (rho,Kstack,Ccode rs) in
  let is =
    (call_stdlib l_print_string (Cvar rho) [Addr usage_msg,[]])@
    (call_stdlib l_exit (Cvar rho) [Immed 2,[]])@
    [Jmp (Addr loop,[])] in
  (usage,c,Array.of_list is)
;;

let uncaught_exn_cb : code_block =
  let rho = newvar () in
  let rs =
    Dict.inserts (Dict.empty compare)
      [Esp,(Capp (Cprim PCstackptr,Cvar rho));Eax,Cprim PCexn] in
  let c =
    Cforall (rho,Kstack,Ccode rs) in
  let is =
    (call_stdlib l_print_string (Cvar rho) [Addr uncaught_exn_msg,[]])@
    (call_stdlib l_exit (Cvar rho) [Immed 1,[]])@
    [Jmp (Addr loop,[])] in
  (uncaught_exn,c,Array.of_list is)
;;

let loop_cb : code_block =
  (loop,Ccode (Dict.empty compare),[|Jmp (Addr loop,[])|])
;;

let s1 = Ccons (Cprim (PCbytes Byte4),Cempty);;
let s2 =
  let rs =
    Dict.inserts (Dict.empty compare)
      [Eax,Cprim PCexn; Esp,Capp (Cprim PCstackptr,Cempty)] in
  Ccons (Ccode rs,Cempty)
;;

let l_tal_main_cb l : code_block =
  let rs =
    Dict.singleton compare Esp (Capp (Cprim PCstackptr,Cempty)) in
  let is =
    [ Alen (Byte4,Eax,Addr l_args);
      Cmp (Reg Eax,Immed 2);
      Jcc (NotEq,(usage,[Tapp Cempty]));
      Mov (Reg Eax,(Immed 1,[]));
      Asub (Byte4,Eax,Addr l_args,Eax) ]@
    ( call_stdlib l_int_of_string Cempty [(Reg Eax,[])] )@
    [ Push (Addr uncaught_exn,[Tapp Cempty]);
      Mov (Reg Ebp,(Reg Esp,[])) ]@
    ( call_kml l s1 Cempty [Reg Eax,[]] )@
    ( call_closure Eax s1 Cempty [] )@
    [ Mov (Reg Eax,(Prjr ((Eax,[]),0),[])) ]@
    ( call_closure Eax Cempty Cempty [] )@
    ( call_stdlib l_print_int s2 [Reg Eax,[]] )@
    ( call_stdlib l_print_newline s2 [] )@
    ( call_stdlib l_exit s2 [Immed 0,[]] )@
    [ Jmp (Addr loop,[]) ] in
  (l_tal_main,Ccode rs,Array.of_list is)
;;

(* The generator *)

let gen_main int_ref entry_l =
  { imports=[|"tal.tali";"kml_types.tali";"stdlib.tali";int_ref|];
    typedefs=[||];
    exports=[|"tal_prog.tali"|];
    con_blocks=[||];
    code_blocks=[|l_tal_main_cb entry_l;usage_cb;uncaught_exn_cb;loop_cb|];
    data_blocks=[|usage_msg_db;uncaught_exn_msg_db|]
  }
;;

(* EOF: cogen_main.ml *)
